#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2022-09-17
# File  : cisco_cellular_lte.py

#
# Monitor status of Cisco cellular modems/connections/interfaces
#
# 2022-09-20: added WATO options
# 2022-11-10: added upper_levels to perfdata for scalars in metrics
# 2022-11-16: added band as metric on special request
#
# sample snmpwalk
#

#
# sample string_table
#

#
# sample section
#

# IMSI -> International Mobile Subscriber Identifier
# IMEI -> International Mobile Equipment Identifier
# ICCID -> Integrated Circuit Card ID
# MSISDN -> Mobile Subscriber Integrated Services Digital Network Number
# FSN -> Factory Serial Number
# RSSI -> Received Signal Strength Indicator
# RSRP -> Reference Signal Received Power
# RSRQ -> Reference Signal Received Quality
# SNR -> Signal to Noise Ratio
#
# for Values https://usatcorp.com/faqs/understanding-lte-signal-strength-values/
#

import time
from dataclasses import dataclass
from typing import Optional, Dict, List
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    StringByteTable,
)

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    SNMPTree,
    contains,
    check_levels,
    OIDEnd,
    OIDBytes,
    Result,
    render,
    State,
    all_of,
    exists,
    get_rate,
    GetRateError,
    get_value_store,
    Metric,
)


###########################################################################
#
#  DATA Parser function
#
###########################################################################

@dataclass
class CiscoCellular3g:
    modem_state: str
    service_type: str
    roaming_status: str
    current_system_time: str
    connection_status: str
    total_bytes_tx: int
    total_bytes_rx: int
    modem_status: str
    network: str
    channel: int
    band: str
    rssi: Optional[int]


@dataclass
class CiscoCellularRadio:
    lte_rsrp: Optional[int]
    lte_rsrq: Optional[int]


@dataclass
class CiscoCellularProfile:
    apn_type: str
    apn: str
    radio_index: str


@dataclass
class CiscoCellularInterface:
    conn_status: str
    pdn_type: str
    ipv4_addr: str
    ipv6_addr: str
    profile: Optional[CiscoCellularProfile]
    radio: Optional[CiscoCellularRadio]
    modem: Optional[CiscoCellular3g]


_cisco_gsm_modem_state = {
    '1': 'Unknown',
    '2': 'Up',
    '3': 'Down',
}

_cisco_gsm_modem_status = {
    '1': 'unknown',
    '2': 'Offline',
    '3': 'Online',
    '4': 'low Power Mode',
}

_cisco_gsm_service_type = {
    '0': '1xRTT',
    '1': 'EVDO Revision 0',
    '2': 'EVDO Revision A',
    '3': 'EVDO Revision B',
    '4': 'GPRS',
    '5': 'EDGE',
    '6': 'UMTS/WCDMA',
    '7': 'HSDPA',
    '8': 'HSUPA',
    '9': 'HSPA',
    '10': 'HSPA Plus',
    '11': 'LTE TDD',
    '12': 'LTE FDD',
}

_cisco_gsm_band = {
    '0': 'UMTS',  # not part of the Cisco MIB
    '1': 'unknown',
    '2': 'invalid',
    '3': 'none',
    '4': 'GSM-850',
    '5': 'GSM-900',
    '6': 'GSM-1800',
    '7': 'GSM-1900',
    '8': 'WCDMA-800',
    '9': 'WCDMA-850',
    '10': 'WCDMA-1900',
    '11': 'WCDMA-2100',
    '12': 'LTE',
}

_cisco_roaming_status = {
    '1': 'Unknown',
    '2': 'Yes',
    '3': 'No',
}

_cisco_gsm_connection_status = {
    '1': 'unknown',
    '2': 'error',
    '3': 'connecting',
    '4': 'dormant',
    '5': 'connected',
    '6': 'disconnected',
    '7': 'idle',
    '8': 'active',
    '9': 'inactive',
}

_cisco_wan_cell_ext_protocol_type = {
    '1': 'unknown',
    '2': 'IPv4',
    '3': 'PPP',
    '4': 'IPv6',
    '5': 'IPv4v6',
}

_cisco_gsm_pdp_status = {
    '1': 'unknown',
    '2': 'active',
    '3': 'inactive',
}


def parse_cisco_cellular_lte(string_table: List[StringByteTable]) -> Optional[Dict[str, CiscoCellularInterface]]:
    section = {}
    radios = {}
    profiles = {}
    modems = {}
    interfaces = {}
    modem_table, radio_table, profile_table, pdn_table, interface_table = string_table

    for radio_oid_end, current_rsrp, current_rsrq in radio_table:
        radios[radio_oid_end] = CiscoCellularRadio(
                lte_rsrp=int(current_rsrp) if current_rsrp.lstrip('-').isdigit() else None,
                lte_rsrq=int(current_rsrq) if current_rsrq.lstrip('-').isdigit() else None,  # // 10,  # See Cisco bug ID CSCvt55347
            )

    for profile_oid_end, apn_type, apn in profile_table:
        radio_index, profile_index = profile_oid_end.split('.')
        profiles[profile_index] = CiscoCellularProfile(
            radio_index=radio_index,
            apn=apn,
            apn_type=apn_type
        )

    for modem_oid_end, modem_state, service_type, roaming_status, modem_time, conn_status,\
         modem_status, network, totyl_bytes_tx, total_bytes_rx, rssi, channel, band in modem_table:

        if band == 1:
            if channel in ['2938', '2961', '2963', '3011', '3027', '3087']:
                band = '0'
            elif channel in ['6200']:
                band = '12'

        modems[modem_oid_end] = CiscoCellular3g(
            modem_state=modem_state,
            service_type=str(service_type[0]*256 + service_type[1]),
            roaming_status=roaming_status,
            current_system_time=modem_time,
            connection_status=conn_status,
            total_bytes_tx=int(totyl_bytes_tx),
            total_bytes_rx=int(total_bytes_rx),
            modem_status=modem_status,
            network=network,
            rssi=int(rssi) if rssi.lstrip('-').isdigit() else None,
            channel=int(channel),
            band=band,
        )

    for index, name in interface_table:
        interfaces[index] = name

    for if_index, pdn_profile_used, pdn_conn_status, pdn_type, pdn_ipv4_addr, pdn_ipv6_addr in pdn_table:
        if interfaces.get(if_index):
            profile = profiles.get(pdn_profile_used)
            if profile:
                radio_index = profile.radio_index
                radio = radios.get(radio_index)
                modem = modems.get(radio_index)
            else:
                radio = None
                modem = None

            section[interfaces.get(if_index)] = CiscoCellularInterface(
                conn_status=pdn_conn_status,
                pdn_type=pdn_type,
                ipv4_addr=pdn_ipv4_addr,
                ipv6_addr=pdn_ipv6_addr,
                profile=profile,
                radio=radio,
                modem=modem,
            )

    return section


###########################################################################
#
#  INVENTORY function
#
###########################################################################


def discovery_cisco_cellular_lte(params, section: Dict[str, CiscoCellularInterface]) -> DiscoveryResult:
    for item, interface in section.items():
        if not params['only_configured'] or interface.profile is not None:
            yield Service(item=item)

###########################################################################
#
#  CHECK function
#
###########################################################################


def check_cisco_cellular_lte(item, params, section: Dict[str, CiscoCellularInterface]) -> Optional[CheckResult]:
    try:
        interface = section[item]
    except KeyError:
        # yield Result(state=State.UNKNOWN, notice='Item not found.')
        return

    if interface.profile is None:
        yield Result(state=State(params['no_profile_state']), summary='No GSM/LTE SIM card associated')
        return

    metric_prefix = 'cisco_cellular_'

    text = f'Status: {_cisco_gsm_pdp_status.get(interface.conn_status)}'
    if interface.conn_status == '2':  # active
        yield Result(state=State.OK, summary=text)
    else:
        yield Result(state=State(params['connection_state']), summary=text)

    text = f'Modem state: {_cisco_gsm_modem_state.get(interface.modem.modem_state)}'
    if interface.modem.modem_state == '2':  # UP
        yield Result(state=State.OK, notice=text)
    else:
        yield Result(state=State(params['modem_not_up_state']), notice=text)

    text = f'Modem status: {_cisco_gsm_modem_status.get(interface.modem.modem_status)}'
    if interface.modem.modem_status == '3':  # online
        yield Result(state=State.OK, notice=text)
    else:
        yield Result(state=State(params['modem_not_online_state']), notice=text)

    yield Result(
        state=State.OK,
        notice=f'Connection status: {_cisco_gsm_connection_status.get(interface.modem.connection_status)}'
    )

    text = f'Roaming: {_cisco_roaming_status.get(interface.modem.roaming_status)}'
    if interface.modem.roaming_status == '3':  # Home not roaming
        yield Result(state=State.OK, notice=text)
    else:
        yield Result(state=State(params['roaming_state']), notice=text)

    if not params['CSCvt55347_fixed'] and interface.radio.lte_rsrq:
        interface.radio.lte_rsrq = interface.radio.lte_rsrq / 10  # handle Cisco BugID CSCvt55347

    if interface.radio.lte_rsrp < -111:
        lte_rsrp_quality = 'Poor'
    elif interface.radio.lte_rsrp <= -103:
        lte_rsrp_quality = 'Fair'
    elif interface.radio.lte_rsrp <= -85:
        lte_rsrp_quality = 'Good'
    elif interface.radio.lte_rsrp >= -84:
        lte_rsrp_quality = 'Excellent'

    if interface.radio.lte_rsrq < -12:
        lte_rsrq_quality = 'Poor'
    elif interface.radio.lte_rsrq <= -9:
        lte_rsrq_quality = 'Fair'
    elif interface.radio.lte_rsrq <= -5:
        lte_rsrq_quality = 'Good'
    elif interface.radio.lte_rsrq > -5:
        lte_rsrq_quality = 'Excellent'

    if interface.modem.rssi < -85:
        rssi_quality = 'Poor'
    elif interface.modem.rssi <= -75:
        rssi_quality = 'Fair'
    elif interface.modem.rssi <= -65:
        rssi_quality = 'Good'
    elif interface.modem.rssi >= -65:
        rssi_quality = 'Excellent'

    for value, label, render_func, levels_lower, metric, in [
        (interface.radio.lte_rsrp, 'RSRP', lambda v: f'{v} dBm ({lte_rsrp_quality})', params['rsrp_levels_lower'], 'rsrp'),
        (interface.radio.lte_rsrq, 'RSRQ', lambda v: f'{v} dB ({lte_rsrq_quality})', params['rsrp_levels_lower'], 'rsrq'),
        (interface.modem.rssi, 'RSSI', lambda v: f'{v} dBm ({rssi_quality})', params['rssi_levels_lower'], 'rssi'),
    ]:
        if value:
            warn, crit = levels_lower
            if warn:
                warn = warn * -1
            if crit:
                crit = crit * -1
            yield from check_levels(
                value=value,
                label=label,
                render_func=render_func,
                levels_upper=(warn, crit),  # used for scalars in metrics
                levels_lower=levels_lower,
                metric_name=f'{metric_prefix}{metric}',
            )

    if interface.radio.lte_rsrp or interface.radio.lte_rsrq:
        if not interface.radio.lte_rsrq or not interface.radio.lte_rsrp:
            yield Result(state=State.OK, notice='Channel changed while polling')

    now_time = time.time()
    value_store = get_value_store()

    for metric, value, label in [
        ('if_in_bps', interface.modem.total_bytes_rx, 'In'),
        ('if_out_bps', interface.modem.total_bytes_tx, 'Out'),
    ]:
        try:
            value = get_rate(value_store, f'{metric}', now_time, value * 8, raise_overflow=True)
        except GetRateError:
            value = 0
        yield from check_levels(
            value=value,
            label=label,
            metric_name=metric,
            render_func=render.networkbandwidth,
        )

    if interface.ipv4_addr:
        yield Result(state=State.OK, notice=f'IPv4 Address: {interface.ipv4_addr}')
    if interface.ipv6_addr:
        yield Result(state=State.OK, notice=f'IPv6 Address: {interface.ipv6_addr}')

    yield Result(state=State.OK, notice=f'Networkcode: {interface.modem.network}')

    expected_band = params['expected_band']
    if interface.modem.band == expected_band:
        yield Result(state=State.OK, summary=f'Band: {_cisco_gsm_band.get(interface.modem.band)}')
    else:
        yield Result(
            state=State(params['state_not_expected_band']),
            summary=f'Band: {_cisco_gsm_band.get(interface.modem.band)} (expected: {_cisco_gsm_band.get(expected_band)})'
        )
    if params['band_as_metric']:
        yield Metric(
            name=f'{metric_prefix}band',
            value=int(interface.modem.band) + 1,
        )

    yield Result(state=State.OK, notice=f'Channel: {interface.modem.channel}')
    yield Result(state=State.OK, notice=f'Service type: {_cisco_gsm_service_type.get(interface.modem.service_type)}')
    yield Result(state=State.OK, notice=f'Provider Time: {interface.modem.current_system_time}')
    yield Result(state=State.OK, notice=f'APN: {interface.profile.apn}')
    yield Result(
        state=State.OK,
        notice=f'APN type: {_cisco_wan_cell_ext_protocol_type.get(interface.pdn_type)}/'
               f'{_cisco_wan_cell_ext_protocol_type.get(interface.pdn_type)} (configured/current)')

###########################################################################
#
#  SNMP section
#
###########################################################################


register.snmp_section(
    name='cisco_cellular_lte',
    parse_function=parse_cisco_cellular_lte,
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.661.1',  # CISCO-WAN-3G-MIB::ciscoWan3gMIBObjects
            oids=[
                OIDEnd(),
                '1.1.3',  # c3gModemState
                OIDBytes('1.1.5'),  # c3gCurrentServiceType
                '1.1.6',  # c3gRoamingStatus
                '1.1.7',  # c3gCurrentSystemTime
                '1.1.8',  # c3gConnectionStatus
                '3.1.1.6',  # c3gModemStatus
                '3.2.1.9',  # c3gGsmNetwork
                '3.2.1.19',  # c3gGsmTotalByteTransmitted
                '3.2.1.20',  # c3gGsmTotalByteReceived
                '3.4.1.1.1',  # c3gCurrentGsmRssi
                '3.4.1.1.4',  # c3gGsmChannelNumber
                '3.4.1.1.3',  # c3gGsmCurrentBand
            ]),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.817.1.1.1.1.1',  # CISCO-WAN-CELL-EXT-MIB::cwceLteRadioEntry
            oids=[
                OIDEnd(),
                '1',  # cwceLteCurrRsrp
                '2',  # cwceLteCurrRsrq
            ]),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.817.1.1.2.3.1',  # CISCO-WAN-CELL-EXT-MIB::cwceLteProfileEntry
            oids=[
                OIDEnd(),
                '2',  # cwceLteProfileType
                '5',  # cwceLteProfileApn
            ]),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.817.1.1.2.4.1',  # CISCO-WAN-CELL-EXT-MIB::cwceLtePdnEntry
            oids=[
                OIDEnd(),
                '2',  # cwceLtePdnProfileUsed
                '3',  # cwceLtePdnConnStatus
                '4',  # cwceLtePdnType
                '5',  # cwceLtePdnIpv4Addr
                '6',  # cwceLtePdnIpv6Addr
            ]),

        SNMPTree(
            base='.1.3.6.1.2.1.31.1.1.1',  # IF-MIB::ifXEntry
            oids=[
                OIDEnd(),
                '1',  # ifName
            ]
        )
    ],
    detect=all_of(
        contains('.1.3.6.1.2.1.1.1.0', 'cisco'),  # sysDescr
        exists('.1.3.6.1.4.1.9.9.661.1.3.1.1.2.*'),  # CISCO-WAN-3G-MIB::c3gImei
    ))

register.check_plugin(
    name='cisco_cellular_lte',
    service_name='Cellular %s',
    discovery_function=discovery_cisco_cellular_lte,
    discovery_ruleset_name='discovery_cisco_cellular_lte',
    discovery_default_parameters={
        'only_configured': True,
    },
    check_function=check_cisco_cellular_lte,
    check_default_parameters={
        'rsrp_levels_lower': (-103, -112),
        'rsrq_levels_lower': (-9, -13),
        'rssi_levels_lower': (-75, -85),
        'roaming_state': 1,
        'no_profile_state': 1,
        'connection_state': 2,
        'modem_not_up_state': 2,
        'modem_not_online_state': 2,
        'expected_band': '12',
        'state_not_expected_band': 1,
        'CSCvt55347_fixed': False,
        'band_as_metric': False
    },
    check_ruleset_name='cisco_cellular_lte',
)
